import os
import sys
import random
import binascii
import unittest

def __b(size):
    def binary_print(num):
        return " ".join(map(
            lambda c: c.zfill(size),  #make sure each chunk is padded with zeros to size
            chunk(format(num, 'b'),-1*size) #chunk into size chunks, starting from the right
            ))
    return binary_print

_b4 = __b(4)
_b8 = __b(8)
_b16 = __b(16)
_b = _b8
_x = lambda b: binascii.hexlify(b, " ", -4)

def example_bytes(length):
    return bytearray(random.getrandbits(8) for _ in range(length))

def demonstrate_chunk():
    ab = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    print(chunk(ab, 5))
    print(chunk(ab, -5))

def chunk(b:bytes, size:int):
    """
    if size is positive, chunks starting left to right
    if size is negative, chunk from the right instead
    chunk size is abs(size)
    chunk size of 0 is an error
    returns a list of chunks of that size
    """
    fromright = size < 0
    size=abs(size)
    if fromright:
        return [ b[::-1][i:i+size][::-1] for i in range(0, len(b), size)][::-1]
        #I'm not sorry
        #okay, maybe a little bit.
    else:
        return [ b[i:i+size] for i in range(0, len(b), size)]

def test_b(x):
    print(_b4(int(x)))
    print(_b(int(x)))
    print(_b8(int(x)))
    print(_b16(int(x)))

class dattr(dict):
    """
    "dict-attr", used for when you don't want to type [""] all the time
    (and i think it looks nicer for things like config settings)
    """
    def __getattr__(self,name):
        """
        With a dattr, you can do this:
        >>> x = dattr({"abc":True})
        >>> x.abc
        True

        """
        if type(self[name]) == type({}): 
            #make sure we wrap any nested dicts when we encounter them
            self[name] = dattr(self[name]) #has to assign to save any changes to nested dattrs
            #e.g.  x.abc.fed = "in" 
        #otherwise just make our key,value pairs accessible through . (e.g. x.name)
        return self[name]
    def __setattr__(self,name,value):
        """
        With a dattr, you can do this:

        >>> x = dattr({"abc":True})
        >>> x.abc = False
        >>> x.abc
        False
        """
        self[name] = value


class test_nested_dattr(unittest.TestCase):
    def test_get(self):
        x = dattr({"abc":{
            "fed":"out"
            }})
        self.assertEqual(x.abc.fed, "out")

    def test_set(self):
        x = dattr({"abc":{
            "fed":"out"
            }})
        x.abc.fed = "in"
        self.assertEqual(x.abc.fed, "in")

def c_array_init_file(filename):
    with open(filename,"rb") as fd:
        c_array_init(fd.read())

def c_array_init(bs:bytes):
    print("uint8_t sample_stream[]={")
    line = ""
    cnt = 0
    for b in bs:
        line += hex(b) 
        line +=","
        cnt += 1
        if cnt == 4:
            line+="\t"
        if cnt >= 8:
            print(line)
            line = ""
            cnt = 0
            continue
    print("}")
    #cat filename |grep -o 'x' |wc -l to know how big it is

if __name__ == "__main__":
    vars()[sys.argv[1]](*sys.argv[2:])




helptext="""
`python -m m17` doesn't do anything on it's own (yet).
Examples:

Without `pip install m17[Codec2]`
    python -m m17.address W2FBI


With `pip install m17[Codec2]`
    python -m m17.audio_test 3200
    python -m m17.client 
"""
def main():
    print(helptext)


if __name__ == "__main__":
    main()




import bitstruct
try:
    from .address import Address
    from .const import *
    from .misc import _x, chunk, example_bytes
except:
    from address import Address
    from const import *
    from misc import _x, chunk, example_bytes

class initialLICH:
    """
    parts that get replicated in regularFrames:
        48b  Address dst
        48b  Address src
        16b  int(M17_streamtype)
        112b nonce (for encryption)
    """
    sz = int((48+48+16+112)/8)
    def __init__(self, 
            src:Address=None, 
            dst:Address=None,
            streamtype=None,
            nonce=None):

        self.src = src
        self.dst = dst
        self.streamtype = streamtype
        self.nonce = nonce 
        assert len(nonce) == 14

    def __eq__(self, other):
        return bytes(self) == bytes(other)
        # for name in ["src","dst","streamtype","nonce"]:
            # x = getattr(self,name)
            # y = getattr(other,name)
            # z = x == y
            # print(z,x,y) 


    def __str__(self):
        return "LICH: " + self.src.callsign + " =[%d]> "%(self.streamtype) + self.dst.callsign

    def __bytes__(self):
        b = b""
        b += bitstruct.pack("u48",self.dst.addr)
        b += bitstruct.pack("u48",self.src.addr)
        b += bitstruct.pack("u16",self.streamtype)
        b += self.nonce
        return b

    def chunks(self):
        me = bytes(self)
        return chunk( me, 6)

    @staticmethod
    def from_bytes(data:bytes):
        d = initialLICH.dict_from_bytes(data)
        return initialLICH(**d)

    @staticmethod
    def dict_from_bytes(data:bytes):
        d = {}
        d["dst"], d["src"], d["streamtype"] = bitstruct.unpack("u48u48u16", data[:14])
        d["dst"] = Address(addr=d["dst"])
        d["src"] = Address(addr=d["src"])
        d["nonce"] = data[14:14+14]
        return d

    @staticmethod
    def recover_bytes_from_bytes_frames( bytes_frames:list):
        frames = [ regularFrame.dict_from_bytes(b) for b in bytes_frames ]
        #frame number gives us the idea of which part of the LICH we have in it
        #assuming that the first regular frame has the first part of the lich
        #per the spec
        #so we really just need one each of the "numbers on the clock"
        #so we could do this better by taking in packets and returning a LICH once 
        #each spot is filled, which might be more robust

        #code below assumes there's a solution in the frames, that there's only one frame
        #for each spot on the "clock", etc, so it will be fragile with real RF

        idx_frame_number = [ (frames.index(x), x["frame_number"]) for x in frames]
        # idx_frame_number = [ (0,9),(1,10),(2,11),(3,12),(4,13) ]
        #reorder so fn%5 == 0 is first
        # print( list(map( lambda x: x, idx_frame_number) ) )
        zeroth_chunk = list(map( lambda x: x[1]%5 == 0, idx_frame_number)).index(True)
        # print(zeroth_chunk)
        reordered = frames[zeroth_chunk:] + frames[0:zeroth_chunk]
        # print(reordered)
        b=b""
        for p in reordered:
            b += p["lich_chunk"]
        return b


class regularFrame:
    """
    48b  LICH chunk
    16b  Frame number counter
    128b payload
    16b  CRC-16 chksum
    """
    sz = int((48+16+128)/8)
    lich_chunk_sz = int(48/8);
    payload_sz = int(128/8)
    def __init__(self, frame_number, payload, LICH:initialLICH=None, lich_chunk:bytes=None):
        """
        Can instantiate with either a full LICH object or just a lich_chunk 
        """
        self.LICH = LICH
        self.lich_chunk = lich_chunk
        self.frame_number = frame_number
        self.payload = payload
        if self.LICH:
            self.LICH_chunks = self.LICH.chunks()
    def __eq__(self, other):
        return bytes(self) == bytes(other)

    def __str__(self):
        return "M17[%d]: %s"%(self.frame_number,_x(self.payload))

    def __bytes__(self):
        b=b""
        if self.LICH:
            lich_chunk_idx = self.frame_number % 5; 
            assert len(self.LICH_chunks[lich_chunk_idx]) == 48/8
            b += self.LICH_chunks[lich_chunk_idx]
        else:
            b += self.lich_chunk
        b += bitstruct.pack("u16", self.frame_number)
        b += bytes(self.payload)
        b += bytes([0]*2) #crc16
        # b += bitstruct.pack("u4",0)
        return b

    @classmethod
    def from_bytes(cls,data:bytes):
        d = cls.dict_from_bytes(data)
        return cls(**d)

    @staticmethod
    def dict_from_bytes(data:bytes):
        d = {}
        d["lich_chunk"] = data[0:6]
        d["frame_number"]= bitstruct.unpack("u16", data[6:8])[0]
        d["payload"] = data[8:8+16]
        #ignore the CRC, it's not implemented yet
        # d["chksum"] = bitstruct.unpack("u16", data[8+16:8+16+2])[0]
        return d

class ipFrame(regularFrame):
    """
    32b "M17 " 
    16b  StreamID
    ?    Full LICH bytes
    16b  Frame number counter
    128b payload
    16b  CRC-16 chksum
    """
    sz = 4+2+initialLICH.sz+2+16+2
    def __init__(self, *args, **kwargs):
        self.streamid = kwargs.pop('streamid',0x0)
        super().__init__(*args,**kwargs)
        if not self.LICH:
            raise(Exception("ipFrames need a full LICH passed"))

    def __str__(self):
        return "SID: %04x\n LICH: "%(self.streamid) + self.LICH.src.callsign + " =[%d]> "%(self.LICH.streamtype) + self.LICH.dst.callsign + "\nM17[%d]: %s"%(self.frame_number,_x(self.payload))

    def __bytes__(self):
        b=b""
        b += b"M17 "
        b += bitstruct.pack("u16", self.streamid)
        b += bytes(self.LICH)
        b += bitstruct.pack("u16", self.frame_number)
        b += bytes(self.payload)
        b += bytes([0]*2) #crc16, TODO
        assert self.sz == len(b)
        return b

    @staticmethod
    def is_m17(data:bytes):
        return data[0:4] == b"M17 "

    @staticmethod
    def dict_from_bytes(data:bytes):
        assert ipFrame.is_m17(data)
        d = {}
        d["streamid"]= bitstruct.unpack("u16", data[4:6])[0]
        lich_start = 6
        lich_end = lich_start+initialLICH.sz
        payload_start = lich_end+2
        payload_end = payload_start+16
        d["LICH"] = initialLICH.from_bytes(data[lich_start:lich_end])
        d["frame_number"]= bitstruct.unpack("u16", data[lich_end:payload_start])[0]
        d["payload"] = data[payload_start:payload_end]
        # d["crc"] = bitstruct.unpack("u16", ...
        return d

def is_LICH( b:bytes ):
    """
    No real way to tell other than size with the implementation in this file
    in RF, they would be the same size
    """
    return len(b) == initialLICH.sz

class M17_Frametype(int):

    """
    low bits v high bits
    stream?
    data?
    voice?
    non-codec2?
    non-3200bps?
    2b: encryption-type
    2b: encryption-subtype
    remaining of 16: reserved
    codec2 3200bps voice stream 00101
    """
    fields = [
            (1, "is_stream"),
            (1, "has_data"),
            (1, "has_voice"),
            (1, "non_codec2"),
            (1, "non_3200bps"),
            (2, "enc_type"),
            (2, "enc_subtype"),
            (7, "reserved"),
            ]




import sys
import time

if __name__ == "__main__":
    try:
        import soundcard as sc
        import pycodec2
        import numpy
        print("Successfully imported pycodec2 and soundcard modules, everything should work.")
        sys.exit(0)
    except Exception as e:
        print(e)
        print("Could not import pycodec2 and soundcard modules, you may want to install extra features by doing `pip install m17[Codec2]` later.`")
        sys.exit(1)
        



try:
    from .address import Address
    from .frames import *
except:
    from address import Address
    from frames import *


class M17_RFFramer:
    def __init__(self, *args, **kwargs):
        self.packet_count = 0
        self.LICH = initialLICH(*args,**kwargs)

    def makeLICH(self):
        return bytes(self.LICH)

    @classmethod
    def fromLICH(cls,data:bytes):
        d = initialLICH.dict_from_bytes( data )
        return cls( **d )

    def payload_stream( self, payload:bytes):
        LICH = self.LICH
        payloads = chunk( payload, regularFrame.payload_sz )
        pkts = []
        for p in payloads:
            if len(p) < regularFrame.payload_sz:
                p = p + b"\x00"*(regularFrame.payload_sz - len(p))
            pkt = regularFrame(LICH=LICH, frame_number=self.packet_count, payload=p)
            self.packet_count+=1
            if self.packet_count >= 2**16:
                self.packet_count = 0
            pkts.append(pkt)
        return pkts

class M17_IPFramer(M17_RFFramer):
    def payload_stream( self, payload:bytes):
        #only difference is which frame we use, ipFrame instead of regularFrame
        LICH = self.LICH
        payloads = chunk( payload, regularFrame.payload_sz )
        pkts = []
        for p in payloads:
            if len(p) < regularFrame.payload_sz:
                p = p + b"\x00"*(regularFrame.payload_sz - len(p))
            pkt = ipFrame(LICH=LICH, frame_number=self.packet_count, payload=p)
            self.packet_count+=1
            if self.packet_count >= 2**16:
                self.packet_count = 0
            pkts.append(pkt)
        return pkts




import unittest
import doctest

from m17 import address, frames, framer, misc
def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(address))
    tests.addTests(doctest.DocTestSuite(frames))
    tests.addTests(doctest.DocTestSuite(framer))
    tests.addTests(doctest.DocTestSuite(misc))
    return tests




import sys
import string
import unittest

import bitstruct

from .misc import _x
callsign_alphabet = " " + string.ascii_uppercase + string.digits + "-/." 
#"." is TBD
# print("Alphabet: %s"%(callsign_alphabet))
# print("len(Alphabet): %d"%(len(callsign_alphabet)))


class test_address(unittest.TestCase):
    def setUp(self):
        self.me = Address(callsign="W2FBI")
        self.ref = Address(callsign="XLX307 D")
    def test_string_compare(self):
        self.assertEqual(self.me, "W2FBI")
    def test_num_compare(self):
        self.assertEqual(self.me, 23178783)
    def test_compare(self):
        me2 = Address(addr=23178783)
        self.assertEqual(self.me, me2)
    def test_not_equal(self):
        me2 = Address(addr=23178784)
        self.assertNotEqual(self.me,me2)
    def test_bytes(self):
        self.assertEqual(bytes(self.me), b'\x00\x00\x01a\xae\x1f')

class Address:
    """
    Call with either "addr" or "callsign" to instantiate, e.g.

    >>> from m17.address import Address
    >>> Address(callsign="W2FBI").addr
    23178783

    You can get the hex version using Python's hex()
    >>> hex( 23178783 )
    '0x161ae1f'

    >>> from m17.address import Address
    >>> Address(addr=23178783).callsign
    'W2FBI'

    You can also use it directly, e.g.
    >>> from m17.address import Address
    >>> Address.encode("W2FBI")
    23178783
    >>> Address.decode(23178783)
    'W2FBI'

    Equality tests work like you might hope:
    >>> Address(callsign="W2FBI") == "W2FBI"
    True
    >>> Address(callsign="W2FBI") == 23178783
    True
    >>> Address(callsign="W2FBI") == Address.encode("W2FBI")
    True
    >>> Address(callsign="W2FBI") == Address(addr=23178783)
    True



    """
    def __init__(self, **kwargs):
        for k,v in kwargs.items():
            if k in ["addr","callsign"]:
                setattr(self,k,v)
        self.callsign = self.callsign.upper() if hasattr(self,"callsign") else self.decode(self.addr)
        self.addr = self.addr if hasattr(self,"addr") else self.encode(self.callsign)

    def __str__(self):
        return "%s == 0x%06x"%(self.callsign,self.addr)
    def __bytes__(self):
        return bitstruct.pack("u48",self.addr)

    def __index__(self):
        return self.addr

    def __eq__(self, compareto):
        if type(compareto) == type(""):
            if compareto.isdigit(): #yeah, gross.
                return int(compareto) == self.addr
            else: 
                return compareto.upper() == self.callsign
        elif type(compareto) == type(1):
            return compareto == self.addr
        elif type(compareto) == type(self):
            return int(self) == int(compareto)
        else:
            return False

    @staticmethod
    def to_dmr_id(something):
        # if no db:
        # url = "https://database.radioid.net/static/users.json"
        # requests.get()
        #if db but not found, _check once_ using https://database.radioid.net/api/dmr/user/?id=3125404
        #return an Address encoded for DMR using database lookup?
        #or jsut the ID as an int?
        ...
    @staticmethod
    def from_dmr_id(dmr_int):
        #return an Address encoded for callsign using dmr database lookup to get callsign
        ...
    def is_dmr_id(self):
        return self.callsign.startswith("D") and self.callsign[1:].isdigit()
    def is_dmr_talkgroup(self):
        return any(
                self.is_brandmeister_tg()
                )
    def is_brandmeister_tg(self):
        return self.callsign.startswith("BM") and self.callsign[1:].isdigit()
    def is_dstar_reflector(self):
        return self.callsign.startswith("REF")

    @staticmethod
    def encode(callsign):
        num = 0
        for char in callsign.upper()[::-1]:
            charidx = callsign_alphabet.index(char)
            num *= 40
            num += charidx
            if num >= 40**9:
                raise(Exception("Invalid callsign"))
        return num
    
    @staticmethod
    def decode(addr):
        num = addr
        if num >= 40**9:
            raise(Exception("Not a callsign"))
        chars = []
        while num > 0:
            idx = int(num%40)
            c =  callsign_alphabet[idx] 
            chars.append(c)
            # print(num,idx,c)
            num //= 40
        callsign = "".join(chars)
        return callsign


def show_help():
    print("""
Provide callsigns on the command line and they will be translated into M17 addresses
    """)
if __name__ == "__main__":
    if len(sys.argv) <= 1:
        show_help()
    else:
        for each in sys.argv[1:]:
            print(Address(callsign=each))



#!/usr/bin/env python
import sys
import time
import queue
import socket
import binascii
import multiprocessing

from .const import *
from .misc import example_bytes,_x,chunk,dattr
from .blocks import *

def default_config(c2_mode):
    c2,conrate,bitframe = codec2setup(c2_mode)
    print("conrate, bitframe = [%d,%d]"%(conrate,bitframe) )

    config = dattr({
        "m17":{
            "dst":"",
            "src":"",
            },
        "vox":{
            "silence_threshold":10, #that's measured in queue packets
            },
        "codec2":{
            "c2":c2,
            "conrate":conrate,
            "bitframe":bitframe,
            },
        })
    return config

def voipsim(host="localhost",src="W2FBI",dst="SP5WWP",mode=3200,port=default_port):
    mode=int(mode) #so we can call modular_client straight from command line
    port=int(port)
    config = default_config(mode)
    audio_sim = zeros( size=config.codec2.conrate, dtype="<h", rate=50)
    tx_chain = [audio_sim, codec2enc, m17frame, tobytes, udp_send((host,port))]
    config.m17.dst = dst
    config.m17.src = src
    print(config)
    modular(config, [tx_chain])


def to_icecast(icecast_url, mode=3200,port=default_port):
    mode=int(mode) #so we can call modular_client straight from command line
    port=int(port)
    rx_chain = [udp_recv(port), m17parse, payload2codec2, codec2dec, ffmpeg(icecast_url)]
    # rx_chain = [udp_recv(port), m17parse, tee('m17'), payload2codec2, codec2dec, ffmpeg(icecast_url)]
    config = default_config(mode)
    modular(config, [rx_chain])

def to_pcm(mode=3200,port=default_port):
    mode=int(mode) #so we can call modular_client straight from command line
    port=int(port)
    rx_chain = [udp_recv(port), m17parse, tee('m17'), payload2codec2, codec2dec, teefile('m17.raw'), null]
    config = default_config(mode)
    modular(config, [rx_chain])

def recv_dump(mode=3200,port=default_port):
    mode=int(mode) #so we can call modular_client straight from command line
    port=int(port)
    rx_chain = [udp_recv(port), teefile("rx"), m17parse, tee('M17'), payload2codec2, teefile('out.c2_3200'),codec2dec, teefile('out.raw'), spkr_audio]
    config = default_config(mode)
    modular(config, [rx_chain])

def voip(host="localhost",src="W2FBI",dst="SP5WWP",mode=3200,port=default_port, voipmode="full"):
    mode=int(mode) #so we can call modular_client straight from command line
    port=int(port)

    #this requires remote host to have port forwarded properly and everything - it doesn't
    # reuse the server socket connection (which would support NAT traversal)

    #this means the tx and rx paths are completely separate, which is,
    # if nothing else, simple to reason about

    tx_chain = [mic_audio, codec2enc, vox, m17frame, tobytes, udp_send((host,port))]
    rx_chain = [udp_recv(port), m17parse, payload2codec2, codec2dec, spkr_audio]
    if voipmode == "tx":
        #disable the rx chain
        #useful for when something's already bound to listening port
        rx_chain = []
    if voipmode == "rx":
        #disable the tx chain
        #useful for monitoring incoming packets without sending anything
        tx_chain = []
    config = default_config(mode)

    config.m17.dst = dst
    config.m17.src = src
    print(config)

    modular(config, [tx_chain, rx_chain])

def _echolink_bridge():
    #this is incomplete, and must remain so until a proper reflector enters the picture
    #it's here mostly to save it for later
    echolink_bridge_in = [udp_recv(55501), chunker_b(640), np_convert("<h"), integer_decimate(2), codec2enc, m17frame, tobytes, udp_send((host,port)) ]
    m17_to_echolink = [
            udp_recv(17000), tee("rx"), 
            m17parse, payload2codec2, codec2dec, 
            integer_interpolate(2),
            udp_send(("pidp8i",55500)) 
            ]
    echolink_bridge_monitor = [
                udp_recv(55501), 
                chunker_b(640), 
                np_convert("<h"), 
                integer_decimate(2), 
                codec2enc, 
                tee("RX"),
                m17frame, tobytes,
                m17parse,payload2codec2, 
                tee("RX"),
                codec2dec, 
                spkr_audio
                ]
    echolink_bridge_out = [
            mic_audio, codec2enc, vox, m17frame, tobytes, 
            m17parse, payload2codec2, codec2dec, 
            integer_interpolate(2),
            udp_send(("pidp8i",55500)) 
            ]
    config = default_config(mode)
    modular(config, [tx_chain, rx_chain])

def m17_to_echolink(port=default_port, echolink_host="localhost",mode=3200, echolink_audio_in_port=55500):
    """
    decode and bridge m17 packets to echolink
    (useful for interopability testing)
    """
    chain = [
            udp_recv(port), 
            m17parse, payload2codec2, codec2dec, 
            integer_interpolate(2), #echolink wants 16k audio
            udp_send((echolink_host,echolink_audio_in_port)) 
            ]
    config = default_config(mode)
    config.verbose = 0
    modular(config, [chain])

def _test_chains_example():
    """
    example playground for testing 
    """
    test_chain = [
            mic_audio, 
            codec2enc, #.02ms of audio per q element at c2.3200 in this part of chain
            # delay(5/.02), #to delay for 5s, divide 5s by the time-length of a q element in this part of chain (which does change)
            # tee("delayed c2bytes: "),
            # teefile("out.m17"),
            # vox, 
            # ptt,
            # m17frame, #.04ms of audio per q element at c2.3200
            # tobytes, 
            # udp_send, 
            # udp_recv, 
            # m17parse, 
            # payload2codec2, #back to .02ms per q el
            codec2dec, 
            # null,
            spkr_audio
            ]
    config = default_config(mode)
    config.verbose = 1
    modular(config, [test_chain])

def modular(config, chains):
    """
    Take in a global configuration, and a list of lists of queue
    processing functions, and hook them up in a chain, each function in
    its own process
    Fantastic for designing, developing, and debugging new features.
    """
    #a chain is a series of small functions that share a queue between each pair
    #each small function is its own process - which is absurd, except this
    #is a testing and development environment, so ease of implementation/modification
    #and modularity is the goal
    #this also means we can meet our latency constraint for writing out
    #to the speakers without any effort, even though our total latency
    #from mic->udp is greater than our deadline. 
    #As long as each function stays under the deadline individually, all we do is add latency from sampled->delivered
    #   (well, as long as we have enough processor cores, but it's current_year, these functions still arent that heavy, and its working excellently given what I needed it to do
    #if a function does get slower than realtime, can I make two in its place writing to the same queues?
    #   as long as i have enough cores still, that seems reasonable - but I'll have to think about it
    """
    queues:
    n -> n2 -> n3 -> n4
    n has no inq
    n4 has no outq
    outq for n is inq for n2, etc
    for each chain:
        0 -> 1 -> 2 -> 3
          0    1     2
        if there's an old outq, inq=
        unless at end of chain, create an outq for each fn, outq=

    """
    modules = {
            "chains":chains,
            "queues":[],
            "processes":[],
            }

    for chainidx,chain in enumerate(modules["chains"]):
        inq = None
        for fnidx,fn in enumerate(chain):
            name = fn.__name__
            if fnidx != len(chain):
                outq = multiprocessing.Queue()
                modules["queues"].append(outq)
            else:
                outq = None
            process = multiprocessing.Process(name="chain_%d/fn_%d/%s"%(chainidx,fnidx,name), target=fn, args=(config, inq, outq))
            modules["processes"].append({
                    "name":name,
                    "inq":inq,
                    "outq":outq,
                    "process":process
                    })
            process.start()
            inq = outq
    try:
        procs = modules['processes']
        while 1:
            if any(not p['process'].is_alive() for p in procs):
                print("lost a client process")
                break
            time.sleep(.05)
        #I can see where this is going to need to change
        #it's fine for now, but a real server will need something different
    except KeyboardInterrupt as e:
        print("Got ^C, ")
    finally:
        print("closing down")
        for proc in procs:
            #messy
            #TODO make a rwlock for indicating shutdown
            proc["process"].terminate()


if __name__ == "__main__":
    vars()[sys.argv[1]](*sys.argv[2:])

"""
Good links I found:
https://www.cloudcity.io/blog/2019/02/27/things-i-wish-they-told-me-about-multiprocessing-in-python/

"""



import unittest
from m17 import address, frames, framer, misc, blocks

def load_tests(loader, standard_tests, pattern):
    """
    misc is small, unconnected things, so tests make more sense inline,
    next to the thing tested.
    This proxies the test loader to that file.
    """
    lm = unittest.defaultTestLoader.loadTestsFromModule
    module_list = [
            misc,
            address,
            blocks,
            ]
    x = unittest.TestSuite([lm(x) for x in module_list])
    return unittest.TestSuite(x)




#!/usr/bin/env python
import setuptools
from setuptools import setup
from os import path

#for reference: https://github.com/navdeep-G/setup.py/blob/master/setup.py
#
NAME = 'm17'
DESCRIPTION = 'M17 radio (and radio-over-IP) protocol implementation. https://github.com/M17-project/'
URL = 'https://git.mmcginty.me/mike/pym17'
EMAIL = 'pyM17@tarxvf.tech'
AUTHOR = 'tarxvf'
REQUIRES_PYTHON = '>=3.5.0'
VERSION = '0.0.11'

REQUIRED=[
        "bitstruct",
        "wheel"
        ]

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()



#look into sourcing VERSION from git tag
#what's this __version__ thing I'm seeing?
setuptools.setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    install_requires=REQUIRED,
    packages=setuptools.find_packages(),
    include_package_data=True,
    extras_require={
        "Codec2":[
            "Cython",
            "numpy",
            "soundcard",
            "samplerate",
            "pycodec2"
            ]
        },
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ]
    )




import unittest

try:
    from address import Address
    from misc import example_bytes
    from frames import initialLICH, regularFrame, ipFrame
except:
    from .address import Address
    from .misc import example_bytes
    from .frames import initialLICH, regularFrame, ipFrame

class test_frame_encodings(unittest.TestCase):
    def test_lich(self):
        lich = initialLICH(
                src=Address(callsign="W2FBI"),
                dst=Address(callsign="SP5WWP"),
                streamtype=5,
                nonce=bytes(example_bytes(14)),
                )
        bl = bytes(lich)
        lich2 = initialLICH.from_bytes(bl)
        assert lich == lich2
    def test_regular_frame(self):
        lich = initialLICH(
                src=Address(callsign="W2FBI"),
                dst=Address(callsign="SP5WWP"),
                streamtype=5,
                nonce=example_bytes(14),
                )
        x = regularFrame(
                LICH=lich,
                frame_number=1,
                payload=example_bytes(16)
                );
        y = bytes(x)
        z = regularFrame.from_bytes(y)
        assert z == x

    def test_ip_frame(self):
        lich = initialLICH(
                src=Address(callsign="W2FBI"),
                dst=Address(callsign="SP5WWP"),
                streamtype=5,
                nonce=example_bytes(14),
                )
        x = ipFrame(
                streamid=0xf00d,
                LICH=lich,
                frame_number=1,
                payload=example_bytes(16)
                );
        y = bytes(x)
        z = ipFrame.from_bytes(y)
        assert z == x



import sys
import time

try:
    import pycodec2
    import numpy
except:
    raise(Exception("Need to install with [Codec2]"))

def audio_test_soundcard(mode):
    import soundcard as sc

    mode = int(mode)
    c2 = pycodec2.Codec2( mode )
    conrate = c2.samples_per_frame()
    bitframe = c2.bits_per_frame()
    pa_rate = 8000

    default_mic = sc.default_microphone()
    default_speaker = sc.default_speaker()
    print(default_mic, default_speaker)
    sc_config= {
            "samplerate": pa_rate,
            "blocksize": conrate
            }
    with default_mic.recorder(**sc_config,channels=1) as mic,\
        default_speaker.player(**sc_config,channels=1) as sp:
        while 1:
            audio = mic.record(numframes=conrate) #.transpose()
            audio = audio.flatten()
            audio = audio * 32767
            audio = audio.astype("<h") 
            c2_bits = c2.encode( audio )
            audio = c2.decode( c2_bits )
            audio = audio.astype("float")
            audio = audio / 32767
            sp.play(audio)

if __name__ == "__main__":
    audio_test_soundcard(int(sys.argv[1]))




import sys
import time
import queue
import socket

from .address import Address
from .frames import ipFrame
from .framer import M17_IPFramer
from .const import *
from .misc import example_bytes,_x,chunk,dattr
from .blocks import *

import numpy

def zeros(size, dtype, rate):
    def fn(config, inq, outq):
        while 1:
            outq.put(numpy.zeros(size, dtype))
            time.sleep(1/rate)
    return fn

def null(config, inq, outq):
    """
    Don't do nuffin.

    (Useful for stopping a q from filling up and preventing further processing)

    (Probably used after tee() or teefile() if you just want to collect
    data with those and not do further processing on the data stream)
    """
    while 1:
        x = inq.get()

def tee(header):
    """
    Print incoming items to screen before putting them on the next q
    bytes get printed as hex strings
    "header" parameter gets printed with each queue element to differentiate multiple tee()s
    named like standard UNIX tee(1)
    """
    def fn(config, inq, outq):
        while 1:
            x = inq.get()
            if type(x) == type(b""):
                print(header, _x(x))
            else:
                print(header, x)
            outq.put(x)
    return fn

def ffmpeg(ffmpeg_url):
    ffmpeg_url_example="icecast://source:m17@m17tester.tarxvf.tech:876/live.ogg"
    def fn(config, inq, outq):
        from subprocess import Popen, PIPE, STDOUT
        p = Popen(
                ["ffmpeg","-re", #ffmpeg, and limit reading rate to native speed so we don't spin a whole core with writing zeros to icecast
                    "-f","s16le","-ar", "8000", "-ac", "1", "-i", "/dev/stdin", #the input options
                    "-ar", "48000", "-ac", "2",
                    "-content_type", "'application/mpeg'",  #only support ogg for now
                    ffmpeg_url]
                , stdout=PIPE, stdin=PIPE, stderr=PIPE)
        while 1:
            try: 
                # audio = inq.get_nowait()
                audio = inq.get()
                p.stdin.write( audio.tobytes() )
            except queue.Empty as e:
                sys.stdout.write('aU')
                audio = numpy.zeros(config.codec2.conrate,dtype="<h")
            sys.stdout.flush()
            # time.sleep(1/50)
    return fn

def teefile(filename):
    """
    Same as tee, except assumes elements coming in are bytes, and writes
    them to a provided filename
    """
    def fn(config, inq, outq):
        with open(filename,"wb") as fd:
            try:
                while 1:
                    x = inq.get()
                    fd.write(x)
                    fd.flush()
                    outq.put(x)
            except:
                fd.close()
    return fn

def throttle(n_per_second):
    """
    Read from the inq and only put elements on the outq at (no more than)
    a specified rate in q elements per second. As "no more than" might
    suggest, this is setting a maximum, not a minimum. Minimum is based
    on your hardware and a number of other factors.
    """
    raise(NotImplementedError) #TODO

def delay(size):
    """
    keep a rolling fifo of size "size", creating a predictable delay 
    (assuming they are coming in at a limited rate)

    See "throttle()" for enforcing a rate limit of elements per unit time.
    """
    def fn(config, inq, outq):
        fifo = []
        while 1:
            x = inq.get()
            fifo.insert(0,x)
            if len(fifo) > size:
                outq.put(fifo.pop())
    return fn

def ptt(config, inq, outq):
    """
    take elements off the inq (like audio frames) and only put them on the outq
    if the "push to talk" evaluates truthy.

    Redo this to take a check function as a parameter and close around
    that instead of using config
    """
    while 1:
        x = inq.get()
        if config.ptt.poll(): #this will check for every single packet 
            outq.put(x)

def vox(config,inq,outq):
    """
    Watch for duplicate incoming q elements, and if there's more than
    a configurable threshold in a row, don't copy further duplicates

    This lets you use a microphone mute as a reverse PTT, among other things
    """
    last = None
    repeat_count = 0
    while 1:
        x = inq.get()
        if x == last:
            repeat_count += 1
        else:
            repeat_count = 0
        last = x
        if repeat_count > config.vox.silence_threshold:
            continue
        outq.put(x)


def mic_audio(config,inq,outq):
    """
    Pull audio off the default microphone in 8k, 1 channel, scale it
    into 16bit shorts since soundcard uses floats, and put them in the outq.

    inq is unconnected here.
    """
    import soundcard as sc
    default_mic = sc.default_microphone()
    # print(default_mic)
    conrate = config.codec2.conrate
    with default_mic.recorder(samplerate=8000, channels=1, blocksize=conrate) as mic:
        while 1:
            audio = mic.record(numframes=conrate)
            #we get "interleaved" floats out of here, in a numpy column ([[][][]])(hence the flatten even though it's a single audio channel)
            audio = audio.flatten() * 32767 #scales from -1,1 to signed 16bit int values
            #rest of the system works in little endian shorts, so we scale it up and convert the type
            audio = audio.astype("<h") 
            outq.put(audio)

def spkr_audio(config,inq,outq):
    """
    play mono (c=1) 8k audio into the speaker
    Accepts 16bit shorts, converts to soundcard expected format

    outq is unconnected here.

    buflen allows for slow computers that cant keep the buffers filled
    within the realtime constraints to still play smooth audio - just
    in stutters of specified length in frames
    """
    import soundcard as sc
    default_speaker = sc.default_speaker()
    # print(default_speaker)
    def silence():
        sp.play(numpy.zeros(config.codec2.conrate))

    with default_speaker.player(samplerate=8000, channels=1) as sp:
        buf = [] #allow for playing chunks of audio even when computer is slow
        buflen = 0 #0 is dont use
        while 1:
            #if we stop receiving audio because someone stops transmitting, 
            #we wont get anything off the queue, so we can't block (hence nowait)
            try: 
                audio = inq.get_nowait()
                #rest of system works in LE signed shorts but soundcard expects floats in and out
                #so convert it back to a float, and scale it back down to an appropriate range (-1,1)
                audio = audio.astype("float")
                audio = audio / 32767
                if buflen:
                    if len(buf) < buflen:
                        buf.append(audio)
                        silence()
                    else:
                        for b in buf:
                            sp.play(b)
                        buf = []
                sp.play(audio)
            except queue.Empty as e:
                #and if we have no data, just play zeros
                silence()


def tobytes(config,inq,outq):
    """
    convert everything passing through to bytes, just what it says on the tin
    """
    while 1:
        outq.put(bytes(inq.get()))


def m17frame(config,inq,outq):
    """
    frame incoming codec2 compressed audio frames into M17 packets
    """
    dst = Address(callsign=config.m17.dst)
    src = Address(callsign=config.m17.src)
    print(dst)
    print(src)
    framer = M17_IPFramer(
            dst=dst,
            src=src,
            streamtype=5, #TODO need to set this based on codec2 settings too to support c2.1600
            nonce=b"\xbe\xef\xf0\x0d" + b"a"*10 ) 
    while 1:
        plen = 16 #TODO grab from the framer itself
        #need 16 bytes for M17 payload, each element on q should be 8 bytes if c2.3200
        #this will fail in funny ways if our c2 payloads dont fit exactly on byte boundaries
        d = inq.get()
        while len(d) < plen:
            d += inq.get()
        #TODO generalize to support other Codec2 sizes, and grab data until enough is here to send
        #TODO payload_stream needs to return packets and any unused data from the buffer to support that functionality
        pkts = framer.payload_stream(d)
        for pkt in pkts:
            outq.put(pkt)

def m17parse(config,inq,outq):
    """
    Parse incoming bytes into M17 ipFrames
    """
    while 1:
        f = ipFrame.from_bytes(inq.get())
        if "verbose" in config and config.verbose:
            print(f)
        outq.put(f)


def payload2codec2(config,inq,outq):
    """
    Pull out an M17 payload and return just the raw Codec2 bytes
    """
    byteframe = int(config.codec2.bitframe/8) #TODO another place where we assume codec2 frames are byte-sized
    while 1:
        x = inq.get()
        for x in chunk(x.payload, byteframe):
            assert len(x) == byteframe
            outq.put(x) 

def codec2setup(mode):
    import pycodec2
    c2 = pycodec2.Codec2( mode )
    conrate = c2.samples_per_frame()
    bitframe = c2.bits_per_frame()
    return [c2,conrate,bitframe]

def codec2enc(config,inq,outq):
    """
    exact opposite of codec2dec

    8khz sample rate, 1 channel (mono audio)
    In: 16bit signed shorts, size varies but expects 160 samples for Codec2 3200
    Out: depends on Codec2 mode, but 3200 bps gives 8 bytes per 160 sample raw frame.
    """
    while 1:
        audio = inq.get()
        c2bits = config.codec2.c2.encode(audio)
        assert len(c2bits) == config.codec2.bitframe/8
        outq.put(c2bits)

def codec2dec(config,inq,outq):
    """
    exact opposite of codec2enc
    In: Depends on Codec2 mode, but expects 8 bytes for Codec2 mode 3200
    Out: 16 bit signed shorts, size varies but 160 samples for Codec2 mode 3200
    8khz sample rate, 1 channel (mono audio)
    """
    while 1:
        c2bits = inq.get()
        audio = config.codec2.c2.decode(c2bits)
        outq.put(audio)

def udp_send(sendto):
    """
    Send incoming bytes to udp (host,port)

    sendto is the standard host,port) tuple like ("localhost",17000)
    """
    def fn(config,inq,outq):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setblocking(False)
        while 1:
            bs = inq.get()
            sock.sendto( bs, sendto )
    return fn

def udp_recv(port):
    """
    Receive UDP datagram payloads as bytes and output them to outq
    Maintains UDP datagram separation on the q
    """
    def fn(config,inq,outq):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.bind(("0.0.0.0", port))
        # sock.setblocking(False)
        while 1:
            x, conn = sock.recvfrom( 1500 )
            #1500 is the maximum packet payload size

            #but what do I do with conn data, anything?
            # print(_x(x))
            outq.put(x)
    return fn

def integer_decimate(i):
    """
    For each incoming list of things, skip some of them, decimating the incoming signal
    e.g. integer_decimate(2) will cut a 16khz sampled audio signal to 8khz
    see integer_interpolate for the reverse
    """
    def fn(config,inq,outq):
        while 1:
            x = inq.get()
            de = x[::i]
            outq.put( x[::i] )
    return fn

def integer_interpolate(i):
    """
    for each incoming list of things, interpolate
    useful for going from 8khz audio to 16khz audio to match expected formats

    starting to look uncomfortably like gnuradio, innit?
    """
    def fn(config,inq,outq):
        import samplerate
        resampler = samplerate.Resampler('sinc_best', channels=1)
        while 1:
            x = inq.get()
            y = resampler.process(x, i )
            z = numpy.array(y, dtype="<h")
            outq.put( z )
    return fn

def chunker_b(size):
    """
    Incoming bytes will get chunked into a particular size for downstream
    nodes that don't do their own buffering
    """
    def fn(config,inq,outq):
        buf = b""
        while 1:
            x = inq.get()
            buf += x
            while len(buf) > size:
                outq.put(buf[:size])
                buf = buf[size:]
    return fn

def np_convert(outtype):
    """
    Use numpy to convert incoming elements to a numpy type
    """
    def fn(config,inq,outq):
        while 1:
            x = inq.get()
            y = numpy.frombuffer(x, outtype)
            outq.put(y)
    return fn

def check_ptt():
    """
    is there a good way to do cross platform ptt, or 
    should i abandon that idea and try to improve vox, or 
    just assume linux?
    """
    if int(time.time()/2) % 2 == 0:
        return True
    else:
        return False
